import sys
import time
from datetime import datetime
import os
import struct

import uhal

from optparse import OptionParser

class FlashSpiInterface():
    """

Class for communication with FLASH memory using SPI interface and IPBus
jb+rjs, may 2015

A Guide for Accessing FLASH memory draft 7

The interface allows a command , address and data to be sent over SPI to the configuration FLASH memory and return any data from the memory. So to do this:

* Put command byte with any address and data into outgoing buffer @ base + 100h and onwards. Buffer is 32bits wide and packs bytes in a Big-Endian way ie starting at the MSB.
* Put number of bytes to transfer (sent + received) into num_words register @ base + 0.
* Toggle the Do_SPI bit bit(0), register @ base + 1.

Bit(0) of the Status register @ base + 2 indicates when the interface is transfering data and will go to zero once all the data is sent. (This is not the same as the FLASH memory busy).

The very first byte returned from the memory has no meaning as this occurs at the same time as the outgoing command is being sent.

The following list shows a subset of commands that the FLASH responds to, copied from its data sheet:

Read Flash      (03h & 3 address bytes)(returns data)
Sector Erase	(D8h & 3 address bytes)
Page Program    (02h & 3 address bytes)(+ data)
Write Enable	(06h)
Read Status Register	(05h)(returns data byte)



Some example commands showing buffer contents:
(Unused bytes are given as 'xx')

Example 1) Read FLASH Device ID codes

Outgoing buffer(0) = 9F xx xx xx
Count = 4
Toggle Do_SPI bit

Return buffer(0) = xx 20 BA 18   (for Xilinx KC705 board)
or
Return buffer(0) = xx 01 20 18   (for FTM. Uses different memory)


Example 2) Read FLASH Memory contents, 256 bytes (= 64 words)

Outgoing buffer(0) = 03 + (3 Address bytes) , ie 03 00 00 00
Count = 260
Toggle Do_SPI bit

Return buffer(0)  = xx xx xx xx
Return buffer(1)  = Bytes 0 - 3
Return buffer(2)  = Bytes 4 - 7
...
Return buffer(64) = Bytes 252 - 255



The contents of the FLASH memory on the KC705 board will look something like:

ff ff ff ff  ff ff ff ff  ff ff ff ff  ff ff ff ff
ff ff ff ff  ff ff ff ff  ff ff ff ff  ff ff ff ff
ff ff ff ff  00 00 00 bb  11 22 00 44  ff ff ff ff
ff ff ff ff  aa 99 55 66  20 00 00 00  30 03 e0 01
00 00 00 0b  ?


In the current set-up, 24bits (as 3 bytes) is used for addressing the memory as bytes, which gives a maximum size of FLASH memory available as 0xffffff = 16 MB.
As we are reading using blocks of memory 256 B (block transfer
of block of 65 words, with first word not used), next page
will start at address 0x000100.

It is possible to start to read from any byte address, but size of
block that can be read is limited by maximum buffer size (above mentioned 65 words (includes command and address))


Example 3) Write Enable

Outgoing buffer(0) = 06 xx xx xx
Count = 1
Toggle Do_SPI bit
Return buffer(0) = xx xx xx xx   - nothing returned


Example 4) Read Status (of FLASH memory)

Outgoing buffer(0) = 05 xx xx xx
Count = 2
Toggle Do_SPI bit
Return buffer(0) = xx St xx xx   - Status returned 2nd byte


A Page (as used for reading and writing) is 256 Bytes, but a Sector (as used for erasing) is 256 Pages or 64k Bytes in size.

How long do these instructions take? A Page program(write) takes about 250 us, an Erase Sector takes about 130 ms.

The "Write Enable" command has to be sent before the FLASH memory can be erased or programmed. (set count = 1)

The FLASH memory page/sector must be erased before it can be programmed. (set count = 1)

The Status register is used to determine when a Program or Erase cycle (internal to FLASH memory)has finished.(set count = 1)
The Read Status command in outgoing buffer(0)= 05 xx xx xx

The return buffer(0), status is in the 2nd byte: xx St xx xx.

Status byte:
Bit7 ?
Bit6 Program Error
Bit5 Erase Error
Bit4 ?
Bit3 ?
Bit2 ?
Bit1 Write enabled. Set by the Write enable command
Bit0 Write in progress = FLASH Busy , during programming or erasing

? = ignore , probably 0
For experimenting with Page Erase and Page Program command, suggest using the top Sector of memory which should be well away from the FPGA configuration data.
If so,
the Sector Erase command word = D8 ff ff 00
and
the Page Program command word = 02 ff ff 00

In this case hardware ignores in address 0xffff00 lower two ff's and erases whole block
starting at address 0xff0000. Erasing in this case means setting all bytes to 0xff (all
bites high).

(Dont forget to send a Write Enable command word before each of these, ie: 06 00 00 00





    """

    def __init__(self, IPBusDevice , name='FLASH_SPI_Ram', memorySize = 0x2000000, sizeOfBuffer=65 , pageSize=256, sectorSize = 65536,
        doArbitration=True,timeoutArbitration=600.):

        """
        Size of EEPROMS we use now:

           eFEX EEPROM     64 MBytes == 0x4000000
           FTM  EEPROM     32 MBytes == 0x2000000

        """
        print "Creating FlashSpiInterface object"
        self.hw_interface = IPBusDevice
        self.mem_name = name
        self.buffer_size = sizeOfBuffer
        self.pageSize = pageSize
        self.nPages      = memorySize/pageSize
        self.sectorSize  = sectorSize
        self.nSectors    = memorySize/sectorSize

        self.waitTime    = 0.001       # standard wait time for RW operations, in sec
#!/usr/bin/python3
        self.timeoutTime = 10.          # Maximum wait for RW operations, in sec

        self.printTiming = False
        self.useArbitration = doArbitration

        if doArbitration:
            self.arbitrationInterface = IpBusArbitrationInterface(self.hw_interface,self.mem_name,timeoutArbitration)

    def tryStartTransaction(self):

        if not self.useArbitration:
            return

        if self.arbitrationInterface.tryStartTransaction():
            return                                                # all went fine

        raise IpBusInterfaceArbitrationBusy("IPBus interface arbitration "+self.mem_name+" shows busy")

    def tryFinishTransaction(self):

        if not self.useArbitration:
            return

        if self.arbitrationInterface.tryFinishTransaction():
            return                                                # all went fine

        raise IpBusInterfaceArbitrationError("IPBus interface arbitration "+self.mem_name+" error!")

    def toggleDoSPIBit(self):

        # flip Do_SPI bit
        self.hw_interface.getNode(self.mem_name+".Command").write(0x1)
        self.hw_interface.dispatch()
        self.hw_interface.getNode(self.mem_name+".Command").write(0x0)
        self.hw_interface.dispatch()

        self.waitTillFWInterfaceReady()    # wait for interface ready

    def writeEnable(self):

        input_list = [0]*self.buffer_size

        input_list[0]=0x06000000       # control command
        self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(input_list)
        self.hw_interface.getNode(self.mem_name+".Num_Words").write(0x1)
        self.hw_interface.dispatch()

        # flip Do_SPI bit
        self.toggleDoSPIBit()


    def readFLASHStatusRegister(self):

        input_list = [0]*self.buffer_size

        input_list[0]=0x05000000       # control command
        self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(input_list)
        self.hw_interface.getNode(self.mem_name+".Num_Words").write(0x2)
        self.hw_interface.dispatch()

        # flip Do_SPI bit
        self.toggleDoSPIBit()

        # read back
        Buffer_readback = self.hw_interface.getNode(self.mem_name+".Incoming").readBlock(self.buffer_size)
        self.hw_interface.dispatch()

#!/usr/bin/python3

        return Buffer_readback[0]             # return status word

    def isFLASHBusy(self):

        statusWord = self.readFLASHStatusRegister()

        if (0x10000 & statusWord):
            return True
        else:
            return False

    def isFLASHProgramError(self):

        statusWord = self.readFLASHStatusRegister()

        if (0x400000 & statusWord):
            return True
        else:
            return False

    def isFLASHEraseError(self):

        statusWord = self.readFLASHStatusRegister()

        if (0x200000 & statusWord):
            return True
        else:
            return False


    def isFWInterfaceBusy(self):

        status_word = self.hw_interface.getNode(self.mem_name+".Status").read()
        self.hw_interface.dispatch()

        if status_word > 0 :
            return True
        else:
            return False


    def waitTillFWInterfaceReady(self):

        time.sleep(self.waitTime)     # always wait at least little bit

        startWait = datetime.now()

        while (self.isFWInterfaceBusy()):       # wait if FW interface to FLASH is still busy

            time.sleep(self.waitTime)

            alreadyWaited = datetime.now() - startWait

            if (alreadyWaited.seconds + alreadyWaited.microseconds/1000000.) > self.timeoutTime:
                print "ERROR: Timeout on FW Interface RW"
                raise IOError

        if self.printTiming:
            alreadyWaited = datetime.now() - startWait
            print " Time spent in waitTillFWInterfaceReady() was " , alreadyWaited.seconds + alreadyWaited.microseconds/1000000.


    def waitTillFLASHReady(self):

        startWait = datetime.now()

        if self.isFLASHProgramError():
            print "ERROR: FLASH Program Error bit set"
            raise IOError

        if self.isFLASHEraseError():
            print "ERROR: FLASH Erase Error bit set"
            raise IOError


        while (self.isFLASHBusy()):

            time.sleep(self.waitTime)

            alreadyWaited = datetime.now() - startWait

            if self.isFLASHProgramError():
                print "ERROR: FLASH Program Error bit set"
                raise IOError

            if self.isFLASHEraseError():
                print "ERROR: FLASH Erase Error bit set"
                raise IOError

            if (alreadyWaited.seconds + alreadyWaited.microseconds/1000000.) > self.timeoutTime:
                print "ERROR: Timeout on FLASH RW"
                raise IOError

        if self.printTiming:
            alreadyWaited = datetime.now() - startWait
            print " Time spent in waitTillFLASHReady() was " , alreadyWaited.seconds + alreadyWaited.microseconds/1000000.

    def readID(self):

        input_list = [0]*self.buffer_size

        # send control command
        input_list[0]=0x9f000000
        self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(input_list)
        self.hw_interface.getNode(self.mem_name+".Num_Words").write(0x4)
        self.hw_interface.dispatch()

        # flip Do_SPI bit
        self.toggleDoSPIBit()

        # read ID word
        Buffer_readback = self.hw_interface.getNode(self.mem_name+".Incoming").readBlock(self.buffer_size)
        self.hw_interface.dispatch()

        if self.printTiming:
            print "FLASH Id code = ", hex(Buffer_readback[0])

        return Buffer_readback[0]


    def setBankAddressRegister(self,bankNumber):

        input_list = [0]*self.buffer_size

        input_list[0] = 0x17000000 + bankNumber*0x10000       # control command

        self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(input_list)
        self.hw_interface.getNode(self.mem_name+".Num_Words").write(0x2)
        self.hw_interface.dispatch()

        # flip Do_SPI bit
        self.toggleDoSPIBit()

#!/usr/bin/python3


    def readMemoryPages(self,nFirstPage,nrOfPages):

        """
        Reads nrOfPages pages from FLASH
        nFirstPage is number of first page (not address!)
        returns list with memory content

        """
        input_list = [0]*self.buffer_size

        memory_content = []

        currentMemoryBank=0
        self.setBankAddressRegister(currentMemoryBank)

        if nFirstPage+nrOfPages > self.nPages:
            print "problem in readMemoryPages"
            raise ValueError

        for iPage in range(nFirstPage,nFirstPage+nrOfPages):

            if self.printTiming:
                print " ... loading page nr. ", hex(iPage)

#!/usr/bin/python3

            full_address = iPage*self.pageSize

            if currentMemoryBank != full_address/0x1000000 :        # shift by 25 bits to determine bank number
                currentMemoryBank = full_address/0x1000000
                self.setBankAddressRegister(currentMemoryBank)

            input_list[0]=0x03000000 + (0xffffff & full_address)    # command plus 24 LSB of address

            self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(input_list)
            self.hw_interface.getNode(self.mem_name+".Num_Words").write(self.pageSize+4)     # 4 bytes in first (non-used) word
            self.hw_interface.dispatch()

#!/usr/bin/python3
            self.toggleDoSPIBit()

            self.waitTillFLASHReady()

#!/usr/bin/python3

            Buffer_readback = self.hw_interface.getNode(self.mem_name+".Incoming").readBlock(self.buffer_size)
            self.hw_interface.dispatch()

#!/usr/bin/python3
            for iii in range(1,len(Buffer_readback)):
                memory_content.append(Buffer_readback[iii])
#!/usr/bin/python3

        return memory_content


    def printMemoryPages(self,nFirstPage,nrOfPages):

        """
        Print nrOfPages pages from FLASH
        nFirstPage is number of first page (not address!)

        """

        if nFirstPage+nrOfPages > self.nPages:
            print "problem in printMemoryPages"
            raise ValueError

        for iPage in range(nFirstPage,nFirstPage+nrOfPages):
            print " ... loading page nr. ", hex(iPage)
            memory_content = self.readMemoryPages(iPage,1)

            for iii in range(len(memory_content)):
                print hex(iii)," ", hex(memory_content[iii])   # print as a func of FLASH address


    def printMemory(self):
        """
        Print full FLASH on screen
        """

        self.printMemoryPages(0,self.nPages)


    def clearMemorySectors(self,nFirstSector,nrOfSectors):

        input_list = [0]*self.buffer_size

        if nFirstSector+nrOfSectors > self.nSectors:
           print "problem in readMemoryPages"
           raise ValueError

        currentMemoryBank=0
        self.setBankAddressRegister(currentMemoryBank)

        for iSector in range(nFirstSector,nFirstSector+nrOfSectors):
            print " ... erasing sector nr. ", hex(iSector)

#!/usr/bin/python3
            full_address = iSector*self.sectorSize

            if currentMemoryBank != full_address/0x1000000 :        # shift by 25 bits to determine bank number
                currentMemoryBank = full_address/0x1000000
                self.setBankAddressRegister(currentMemoryBank)

            self.writeEnable()

            input_list[0]=0xd8000000 + (0xffffff & full_address)

            self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(input_list)
            self.hw_interface.getNode(self.mem_name+".Num_Words").write(4)
            self.hw_interface.dispatch()

#!/usr/bin/python3
            self.toggleDoSPIBit()

            self.waitTillFLASHReady()


    def clearMemory(self):

        self.clearMemorySectors(0,self.nSectors)

    def writeMemorySectors(self,nFirstSector,nrOfSectors,data):

        if len(data) != nrOfSectors*self.sectorSize/4:             # always 4 bytes in one write word
            print "problem with data vector in writeMemorySectors", len(data)
            raise ValueError

        input_list = [0]*self.buffer_size

        self.clearMemorySectors(nFirstSector,nrOfSectors)

        currentMemoryBank=0
        self.setBankAddressRegister(currentMemoryBank)

        for iSector in range(nFirstSector,nFirstSector+nrOfSectors):
            print " ... writing into sector nr. ", hex(iSector)

            for iPage in range(self.sectorSize/self.pageSize):          # loop over pages in each sector

                if self.printTiming:
                    print " ... writing page nr. ", hex(iPage), " within this sector"

                for iWord in range(self.pageSize/4):
                    input_list[iWord+1] = data[(iSector-nFirstSector)*self.sectorSize/4+iPage*self.pageSize/4+iWord]

#!/usr/bin/python3

                full_address = iSector*self.sectorSize + iPage*self.pageSize

                if currentMemoryBank != full_address/0x1000000 :        # shift by 25 bits to determine bank number
                    currentMemoryBank = full_address/0x1000000
                    self.setBankAddressRegister(currentMemoryBank)

                input_list[0]=0x02000000 + (0xffffff & full_address)

                self.writeEnable()

                self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(input_list)
                self.hw_interface.getNode(self.mem_name+".Num_Words").write(self.pageSize+4)
                self.hw_interface.dispatch()

#!/usr/bin/python3
                self.toggleDoSPIBit()

                self.waitTillFLASHReady()


    def writeMemory(self,data):

        self.writeMemorySectors(0,self.nSectors,data)


    def writeMemoryFromFile(self,nFirstSector,file_object):

        byte = file_object.read(1)
        byte_nr = 1

        word = 0
        word_nr = 0

        data_to_write = []

        while byte != '':

            byte_unpacked, = struct.unpack('B',byte)

            word = (word << 8) + byte_unpacked

            if byte_nr%4 == 0:
                word_nr += 1
                data_to_write.append(word)
                word = 0

            byte = file_object.read(1)
            byte_nr += 1

#!/usr/bin/python3

        byte_nr -= 1                    # last byte was empty

        if not (byte_nr%4 == 0):
            word = (word << 8*(4-byte_nr%4))
            data_to_write.append(word)

        if self.printTiming:
            for iii in data_to_write:
                print hex(iii)

        n_sectors_data = int(len(data_to_write)/(self.sectorSize/4)+1)
        n_zeroes_toAdd = n_sectors_data * (self.sectorSize/4) - len(data_to_write)

        for iii in range(n_zeroes_toAdd):
            data_to_write.append(0x0)

#!/usr/bin/python3

        self.writeMemorySectors(nFirstSector,n_sectors_data,data_to_write)


    def validateMemoryWithFile(self,nFirstSector,file_object):

#!/usr/bin/python3

        byte = file_object.read(1)
        byte_nr = 1

        word = 0
        word_nr = 0

        data_to_compare = []

        while byte != '':

            byte_unpacked, = struct.unpack('B',byte)

            word = (word << 8) + byte_unpacked

            if byte_nr%4 == 0:
                word_nr += 1
                data_to_compare.append(word)
                word = 0

            byte = file_object.read(1)
            byte_nr += 1

#!/usr/bin/python3

        byte_nr -= 1                    # last byte was empty

        if not (byte_nr%4 == 0):
            word = (word << 8*(4-byte_nr%4))
            data_to_compare.append(word)


        if self.printTiming:
            print " Check of data to be compared with  memory content ;-) "
            for iii in data_to_compare:
                print hex(iii)

        n_sectors_data = int(len(data_to_compare)/(self.sectorSize/4)+1)
        n_zeroes_toAdd = n_sectors_data * (self.sectorSize/4) - len(data_to_compare)

        if self.printTiming:
            print "nr of sectors ",  n_sectors_data , " zeroes to add", n_zeroes_toAdd

        for iii in range(n_zeroes_toAdd):
            data_to_compare.append(0x0)

        firstPageToRead = nFirstSector*(self.sectorSize/self.pageSize)
        nfOfPagesToRead = n_sectors_data*(self.sectorSize/self.pageSize)

        if self.printTiming:
            print "first page to read", firstPageToRead, " nr. of pages to read ", nfOfPagesToRead

        data_from_FLASH = self.readMemoryPages(firstPageToRead,nfOfPagesToRead)

        differences = [i for i, j in zip(data_from_FLASH, data_to_compare) if not i == j]

#!/usr/bin/python3
#!/usr/bin/python3

        for iii in range(len(differences)):
            print hex(differences[iii])

        if len(differences) :
            print "Files are different!"
            return False
        else:
            print "Files are the same!"
            return True


class PLLSpiInterface():
    """
 Class for communication with on-board PLLs using SPI interface and IPBus
 jb+rjs, nov 2015

  This class allows to set PLL registers from file (nine numbers per one PLL, corresponding
  to nine registers in a PLL),
  read value of a register from PLL and 'burn PLL registers', that means store current version
  of PLL registers into PLL eeprom

  Loading PLL:
        - Set the Select register (only need to do this once while accessing the same PLL)
        - Set the count to 9
        - Hand fill the 9 outgoing memory locations as read from the file
        - Toggle the 'do-it' bit(0) of the command register.

  Reading value from PLL register:
        - Set the Select register (only need to do this once while accessing the same PLL)
        - Set the count to 2
        - In the first location (the command) put 0xe in the first four lsbs and the register number in the next four.
          The second location should be all 0s. ie to read register 1 , set the outgoing memory with
           0x0000001e  0x00000000.
        - Toggle the 'do-it' bit. The register value will be returned in the second word of the incoming memory. Ignore
          the first word.

 Burning PLL register values:
        - setup registers using above mentioned method
        - Set the Select register (only need to do this once while accessing the same PLL)
        - Send one word of 0x1F to the PLL


    """
    def __init__(self, IPBusDevice , name='PLL_SPI_Ram', sizeOfBuffer=9,doArbitration=True,timeoutArbitration=1.):

        print "Creating PLLSpiInterface object"
        self.hw_interface = IPBusDevice
        self.mem_name = name
        self.buffer_size = sizeOfBuffer

        self.waitTime    = 0.001       # standard wait time for RW operations, in sec
        self.timeoutTime = 1.          # Maximum wait for RW operations, in sec

        self.printTiming = False
        self.useArbitration = doArbitration

        if doArbitration:
            self.arbitrationInterface = IpBusArbitrationInterface(self.hw_interface,self.mem_name,timeoutArbitration)

    def tryStartTransaction(self):

        if not self.useArbitration:
            return

        if self.arbitrationInterface.tryStartTransaction():
            return                                                # all went fine

        raise IpBusInterfaceArbitrationBusy("IPBus interface arbitration "+self.mem_name+" shows busy")

    def tryFinishTransaction(self):

        if not self.useArbitration:
            return

        if self.arbitrationInterface.tryFinishTransaction():
            return                                                # all went fine

        raise IpBusInterfaceArbitrationError("IPBus interface arbitration "+self.mem_name+" error!")


    def toggleDoSPIBit(self):

        # flip Do_SPI bit
        self.hw_interface.getNode(self.mem_name+".Command").write(0x1)
        self.hw_interface.dispatch()
        self.hw_interface.getNode(self.mem_name+".Command").write(0x0)
        self.hw_interface.dispatch()

        self.waitTillFWInterfaceReady()    # wait for interface ready

    def isFWInterfaceBusy(self):

        status_word = self.hw_interface.getNode(self.mem_name+".Status").read()
        self.hw_interface.dispatch()

        if status_word > 0 :
            return True
        else:
            return False

    def waitTillFWInterfaceReady(self):

        time.sleep(self.waitTime)     # always wait at least little bit

        startWait = datetime.now()

        while (self.isFWInterfaceBusy()):       # wait if FW interface to FLASH is still busy

            time.sleep(self.waitTime)

            alreadyWaited = datetime.now() - startWait

            if (alreadyWaited.seconds + alreadyWaited.microseconds/1000000.) > self.timeoutTime:
                print "ERROR: Timeout on FW Interface RW"
                raise IOError


        if self.printTiming:
            alreadyWaited = datetime.now() - startWait
            print " Time spent in waitTillFWInterfaceReady() was " , alreadyWaited.seconds + alreadyWaited.microseconds/1000000.


    def setupPLL(self,nrOfPLL,inputFile):

        print " setting PLL nr.", nrOfPLL , " from file" , inputFile

#!/usr/bin/python3
        self.hw_interface.getNode(self.mem_name+".Select").write(nrOfPLL)
        self.hw_interface.dispatch()

        input_file = open(inputFile,'r')


        setup_list = []

        input_file.readline()                           # get rid of first line

        for iii in range(self.buffer_size):
            my_line = input_file.readline()
            my_line = my_line.split("\t")
            setup_list.append(int(my_line[1],16))

        self.hw_interface.getNode(self.mem_name+".Num_Words").write(self.buffer_size)
        self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(setup_list)
        self.hw_interface.dispatch()

        self.toggleDoSPIBit()

    def setupPLLEditedFile(self,nrOfPLL,inputFile):

        print " setting PLL nr.", nrOfPLL , " from trimmed file" , inputFile

#!/usr/bin/python3
        self.hw_interface.getNode(self.mem_name+".Select").write(nrOfPLL)
        self.hw_interface.dispatch()

        input_file = open(inputFile,'r')


        setup_list = []

        for iii in range(self.buffer_size):
            setup_list.append(int(input_file.readline(),16))

        self.hw_interface.getNode(self.mem_name+".Num_Words").write(self.buffer_size)
        self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(setup_list)
        self.hw_interface.dispatch()

        self.toggleDoSPIBit()

    def readPLLRegister(self,nrOfPLL,nrOfRegister):

#!/usr/bin/python3
        self.hw_interface.getNode(self.mem_name+".Select").write(nrOfPLL)
        self.hw_interface.dispatch()

#!/usr/bin/python3
        self.hw_interface.getNode(self.mem_name+".Num_Words").write(2)
        command_list=[0xe+0x10*nrOfRegister,0]
        self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(command_list)
        self.hw_interface.dispatch()

        self.toggleDoSPIBit()

        Buffer_readback = self.hw_interface.getNode(self.mem_name+".Incoming").readBlock(2)
        self.hw_interface.dispatch()

        return Buffer_readback[1]

    def readAllPLLRegisters(self,nrOfPLL):

        print "Reading content of registers in PLL nr. ", nrOfPLL

        for iii in range(self.buffer_size):
            print " Rg. nr. ", iii , "    " , hex(self.readPLLRegister(nrOfPLL,iii))


    def burnPLLRegisters(self,nrOfPLL):

#!/usr/bin/python3
        self.hw_interface.getNode(self.mem_name+".Select").write(nrOfPLL)
        self.hw_interface.dispatch()

#!/usr/bin/python3
        self.hw_interface.getNode(self.mem_name+".Num_Words").write(1)
        command_list=[0x1f]
        self.hw_interface.getNode(self.mem_name+".Outgoing").writeBlock(command_list)
        self.hw_interface.dispatch()

        self.toggleDoSPIBit()


class XADCSlowControlInterface():
    """
 Class for communication with on-board FADC via direct IPBus register access
 jb+rjs, oct 2015

    """
    def __init__(self, IPBusDevice , name='XADC'):

        print "Creating XADCSlowControlInterface object"
        self.hw_interface = IPBusDevice
        self.interface_name = name

    def getTemperature(self):

        temperature = self.hw_interface.getNode(self.interface_name+".Temperature").read()
        self.hw_interface.dispatch()
        return 0.00769*int(temperature) - 273.15   # in deg Celsius

    def getVCC_Aux(self):

        voltage = self.hw_interface.getNode(self.interface_name+".VCCAux").read()
        self.hw_interface.dispatch()
        return 3.*int(voltage)/0xffff              # in volts

    def getVCC_Int(self):

        voltage = self.hw_interface.getNode(self.interface_name+".VCCInt").read()
        self.hw_interface.dispatch()
        return 3.*int(voltage)/0xffff              # in volts

    def getVCC_Ram(self):

        voltage = self.hw_interface.getNode(self.interface_name+".VCCBram").read()
        self.hw_interface.dispatch()
        return 3.*int(voltage)/0xffff              # in volts


class ModuleIDInterface():
    """
    Class for communication with registers holding ID info of FPGAS via direct IPBus register access
    Alex Stabekis, 02/2020

    """
    def __init__(self, IPBusDevice , name='Common_IdVersion'):

        print "Creating ID object"
        self.hw_interface = IPBusDevice
        self.interface_name = name

    def getModule_ID(self):

        MyValues=[]

        v0 = self.hw_interface.getNode(self.interface_name+".Module_ID").read()
        v1 = self.hw_interface.getNode(self.interface_name+".Module_ID.L1CaloNumber").read()
        v2 = self.hw_interface.getNode(self.interface_name+".Module_ID.Serial").read()
        v3 = self.hw_interface.getNode(self.interface_name+".Module_ID.Issue").read()
        v4 = self.hw_interface.getNode(self.interface_name+".Module_ID.pcbUpdate").read()
        self.hw_interface.dispatch()

        MyValues.append(v0)
        MyValues.append(v1)
        MyValues.append(v2)
        MyValues.append(v3)
        MyValues.append(v4)
        return MyValues

    def getXmlVersion(self):
        XmlV = self.hw_interface.getNode(self.interface_name+".XmlVersion.Version").read()
        self.hw_interface.dispatch()
        return (XmlV)

    def getFirmwareVersion(self):
        FV = self.hw_interface.getNode(self.interface_name+".FirmwareVersion").read()
        self.hw_interface.dispatch()
        return (FV)


class I2CInterface():
    """
    Generic class for communication wia I2C bus
    jb+rjs, apr 2016 , updated aug 2018

    bit definitions for cmd_stat (8 bits):       START STOP RD WR        ACK/NACK  * * *

    """
    def __init__(self, IPBusDevice , name='I2C_EEPROM', wait_after_STOP = 0.05,frequency=100,doArbitration=True,timeoutArbitration=0.1):

        print "Creating I2CInterface object"
        self.hw_interface = IPBusDevice
        self.slave_name = name

        self.waitTime    = 0.001       # standard wait time for RW operations, in sec
        self.timeoutTime = 1.          # Maximum wait for RW operations, in sec

        self.communication_frequency = frequency     # in kHz

        self.waitAfterSTOP= wait_after_STOP   # wait after STOP for byte write, typically 0.1 for MPODs, 0.005 for FLASH
        self.printTiming = False
        self.doPrinting  = True

        self.useArbitration = doArbitration

        if doArbitration:
            self.arbitrationInterface = IpBusArbitrationInterface(self.hw_interface,self.slave_name,timeoutArbitration)

        self.setFrequency()
        self.enableCore()

    def tryStartTransaction(self):

        if not self.useArbitration:
            return

        if self.arbitrationInterface.tryStartTransaction():
            return                                                # all went fine

        raise IpBusInterfaceArbitrationBusy("IPBus interface arbitration "+self.slave_name+" shows busy")

    def tryFinishTransaction(self):

        if not self.useArbitration:
            return

        if self.arbitrationInterface.tryFinishTransaction():
            return                                                # all went fine

        raise IpBusInterfaceArbitrationError("IPBus interface arbitration "+self.slave_name+" error!")


    def setFrequency(self):

        parameter_value = 32000/(5*self.communication_frequency) - 1

#!/usr/bin/python3

        self.hw_interface.getNode(self.slave_name+".ps_lo").write(parameter_value)
        self.hw_interface.getNode(self.slave_name+".ps_hi").write(0x00)
        self.hw_interface.dispatch()

    def enableCore(self):

        self.hw_interface.getNode(self.slave_name+".ctrl").write(0x80)
        self.hw_interface.dispatch()

    def isTIPOn(self):

        status_word = self.hw_interface.getNode(self.slave_name+".cmd_stat").read()
        self.hw_interface.dispatch()

        if status_word & 0x2 > 0 :         # TIP is second bit (bit nr. 1), if on, need to wait
            return True
        else:
            return False

    def noAckReceived(self):

        status_word = self.hw_interface.getNode(self.slave_name+".cmd_stat").read()
        self.hw_interface.dispatch()

        if status_word & 0x80 > 0 :         # Ack is eight bit (bit nr. 7), if on, Ack was not received
            return True
        else:
            return False



    def waitForTIP(self):

        time.sleep(self.waitTime)     # always wait at least little bit

        startWait = datetime.now()

        while (self.isTIPOn()):       # wait if TIP flag is still on

            time.sleep(self.waitTime)

            alreadyWaited = datetime.now() - startWait

            if (alreadyWaited.seconds + alreadyWaited.microseconds/1000000.) > self.timeoutTime:

                if self.doPrinting:
                    print "ERROR: Timeout on FW Interface RW"
                raise IOError


        if self.printTiming:
            alreadyWaited = datetime.now() - startWait
            print " Time spent in waitForTIP() was " , alreadyWaited.seconds + alreadyWaited.microseconds/1000000.


    def writeByteToBridge(self,deviceAddress,byteData):

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2)  # address of the device shifted by 1 bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)         # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()
        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR from writeByteToBridge() in interface "+self.slave_name+": No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(byteData)         # write byteData
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x50)         # write to slave bit + STOP
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            if self.doPrinting:
                print "ERROR from writeByteToBridge() in interface"+self.slave_name+": No Ack in I2C communication"
            raise IOError

        time.sleep(self.waitAfterSTOP)                                             ### wait after STOP signal


    def writeByte(self,deviceAddress,byteAddress,byteData):

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2)  # address of the device shifted by 1 bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)         # write to slave bit+START

        self.hw_interface.dispatch()

        self.waitForTIP()
        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress)      # address inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()
        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(byteData)         # write value byteData into memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x50)         # write to slave bit + STOP
        self.hw_interface.dispatch()

        self.waitForTIP()
        if self.noAckReceived():

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        time.sleep(self.waitAfterSTOP)                                             ### wait after STOP signal

    def writeBlock(self,deviceAddress,byteAddress,byteData_list):

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2)  # address of the device shifted by 1 bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)         # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()
        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress)      # address inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()
        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        for byteContent in byteData_list:
#!/usr/bin/python3
            self.hw_interface.getNode(self.slave_name+".data").write(byteContent)      # write value byteData into memory
            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
            self.hw_interface.dispatch()

            self.waitForTIP()
            if self.noAckReceived():

                self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
                self.hw_interface.dispatch()

                if self.doPrinting:
                    print "ERROR: No Ack in I2C communication"
                raise IOError

        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP
        self.hw_interface.dispatch()

        time.sleep(self.waitAfterSTOP)   ### wait after STOP signal



    def readByte(self,deviceAddress,byteAddress):

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2)  # address of the device shifted by 1 bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)         # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress)      # address inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError


        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2+1)  # address of the device shifted by 1 bit + read bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)           # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()


            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError


        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x68)           #  set READ , NACK and STOP

        self.hw_interface.dispatch()

        self.waitForTIP()        # don't receive slave Ack now

        memory_content = self.hw_interface.getNode(self.slave_name+".data").read()     # read received byte
        self.hw_interface.dispatch()

        return memory_content

    def readByteFromBridge(self,deviceAddress):

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2+1)  # address of the device shifted by 1 bit + read bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)           # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR from readByteFromBridge() in interface "+self.slave_name+": No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x68)              # READ+ NACK +STOP
        self.hw_interface.dispatch()
        self.waitForTIP()

        memory_content = self.hw_interface.getNode(self.slave_name+".data").read()     # read received byte
        self.hw_interface.dispatch()

        return memory_content


    def readBlock(self,deviceAddress,byteAddress,nBytes):

        outputData = []

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2)  # address of the device shifted by 1 bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)         # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError


        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress)      # address inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError


        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2+1)  # address of the device shifted by 1 bit + read bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)           # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError


       # loop over memory

        for iByte in range(nBytes-1):

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x20)           #  set READ and ACK
            self.hw_interface.dispatch()
            self.waitForTIP()

            memory_content = self.hw_interface.getNode(self.slave_name+".data").read()     # read received byte
            self.hw_interface.dispatch()
            outputData.append(memory_content)

        # now read the last byte

        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x68)           #  set READ , NACK and STOP

        self.hw_interface.dispatch()

        self.waitForTIP()        # don't receive slave Ack now

        memory_content = self.hw_interface.getNode(self.slave_name+".data").read()     # read received byte
        self.hw_interface.dispatch()


        outputData.append(memory_content)

        return outputData

    def writeByteEEPROM(self,deviceAddress,byteAddress0,byteAddress1,byteData):

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2)  # address of the device shifted by 1 bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)         # write to slave bit+START

        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError


        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress0)      # address0 inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress1)      # address1 inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError


        self.hw_interface.getNode(self.slave_name+".data").write(byteData)         # write value byteData into memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x50)         # write to slave bit + STOP
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError


        time.sleep(self.waitAfterSTOP)                                             ### wait after STOP signal


    def readByteEEPROM(self,deviceAddress,byteAddress0,byteAddress1):

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2)  # address of the device shifted by 1 bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)         # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError


        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress0)      # address inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress1)      # address inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2+1)  # address of the device shifted by 1 bit + read bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)           # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError



        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x68)           #  set READ , NACK and STOP

        self.hw_interface.dispatch()

        self.waitForTIP()        # don't receive slave Ack now

        memory_content = self.hw_interface.getNode(self.slave_name+".data").read()     # read received byte
        self.hw_interface.dispatch()

        return memory_content

    def writeBlockEEPROM(self,deviceAddress,byteAddress0,byteAddress1,byteData_list):

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2)  # address of the device shifted by 1 bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)         # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError


        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress0)      # address0 inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress1)      # address1 inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError



        for byteContent in byteData_list:
            print "Writing ", byteContent
            self.hw_interface.getNode(self.slave_name+".data").write(byteContent)      # write value byteData into memory
            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
            self.hw_interface.dispatch()

            self.waitForTIP()

            if self.noAckReceived():

                self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
                self.hw_interface.dispatch()

                if self.doPrinting:
                    print "ERROR: No Ack in I2C communication"
                raise IOError


        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP
        self.hw_interface.dispatch()

        time.sleep(self.waitAfterSTOP)   ### wait after STOP signal


    def readBlockEEPROM(self,deviceAddress,byteAddress0,byteAddress1,nBytes):

        outputData = []

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2)  # address of the device shifted by 1 bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)         # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError



        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress0)      # address inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(byteAddress1)      # address inside memory
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x10)         # write to slave bit
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2+1)  # address of the device shifted by 1 bit + read bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)           # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError



      # loop over memory

        for iByte in range(nBytes-1):

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x20)           #  set READ and ACK
            self.hw_interface.dispatch()
            self.waitForTIP()

            memory_content = self.hw_interface.getNode(self.slave_name+".data").read()     # read received byte
            self.hw_interface.dispatch()
            outputData.append(memory_content)

        # now read the last byte

        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x68)           #  set READ , NACK and STOP

        self.hw_interface.dispatch()

        self.waitForTIP()        # don't receive slave Ack now

        memory_content = self.hw_interface.getNode(self.slave_name+".data").read()     # read received byte
        self.hw_interface.dispatch()

        outputData.append(memory_content)

        return outputData



    def doesBusRespond(self,deviceAddress):

        self.hw_interface.getNode(self.slave_name+".data").write(deviceAddress*2+1)  # address of the device shifted by 1 bit + read bit
        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x90)           # write to slave bit+START
        self.hw_interface.dispatch()

        self.waitForTIP()

        if self.noAckReceived():

            self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x40)         #  STOP signal to release bus
            self.hw_interface.dispatch()
            time.sleep(self.waitAfterSTOP)   ### wait after STOP signal

            if self.doPrinting:
                print "ERROR: No Ack in I2C communication"
            raise IOError

        self.hw_interface.getNode(self.slave_name+".cmd_stat").write(0x68)         #  READ+STOP+NACK signal to release bus
        self.hw_interface.dispatch()
        time.sleep(self.waitAfterSTOP)   ### wait after STOP signal

        return True                                                                 # looks like bus responded!


class IpBusArbitrationInterface():
    """
     Generic class that implements IPBus arbitration mechanism
     jb+rjs, 23 March 2018

     two R/W register per end point to arbitrate access and implement timeout

     Key-lock register:
     bits 31 downto 2: access key
     bit 1           : lock bit
     bit 0           : timeout error bit

     Timeout register:
     bits 31 downto 0 : timeout in ms


     Corresponding piece of address map:

     <node id="Key_Lock" address="0x8">
            <node id="Key"  mask="0x7FFFFFFC"/>
            <node id="Lock" mask="0x2"/>
            <node id="Terr" mask="0x1"/>
       </node>
       <node id="Timeout"  address="0x9"/>



 What is client (SW) doing:

 - request access:
   1. Set the lock bit. Issue RMW transaction setting the lock bit on.
      - if the bit that is read back is zero, continue. On the transition of lock bit from zero to one the FW
        will automatically start the time with some default delay value
      - if the bit that is read back is one, another client has the lock. Throw exception
   2. read access key, remember it
   3. set appropriate time into timeout register, this is time the software interface thinks it will need. Write
      operation on this register automatically restarts the timer


 - during acces:
   1. check (from time to time) that the access key is still the same, if not stop the operation. This needs to be done
      whenever the software is waiting for some condition to happen (for example EEPROM erase to finish). In such cases
      there is SW timeout that needs to be shorter than the timeout set in FW, but to make the procedure safer, the software
      can check that the access key is still the same before resuming operation.

 - after the access:
   1. check the access key is still the same, in not something went wrong
   2. clear the lock bit


 What is FW doing:
 - implements two arbitration registers for each sensitive ipBus endpoint
 - when lock bit goes from one to zero increase the access key (take care of wrap around)
 - when the lock bit goes from zero to one sets the  timeout register to some default value and starts the timer.
   Current default timeout value is 0.25 s (March 2018)

 - implements timeout watchdog. This counter is counting down and reset when there is an write operation on
   the timeout register.
   When the timer times out and the lock bit is set, clear the lock bit. This automatically triggers the increase in the access key.
   Set the protocol error bit. This bit can be read out and monitored by software, to check if there were any
   timeouts. The protocol error bit has to be cleared in software.


 Possible problem situations:

- if software keeps accessing the interface after time-out it may cause problems. This can happen if timeout time set up by
  SW at the beginning is not sufficient. We need to check access key always at the end of SW access to make sure that nothing
  bad happened in the mean time (like FW timeout and other process accessing the interface). In addition we need to
  check the access key always when the process waits (polls) for something, before it resumes the operation.

- sofware may block interface by repeatedly requesting access (by setting lock). It is difficult to protect against
  this, good that our software behaves!

    """
    def __init__(self, IPBusDevice , name='', timeout_to_set = 0.1):

        print "Creating IpBusArbitrationInterface object"

        self.accessKey = None
        self.fwTimeout = int(timeout_to_set*1000)     # time internally in  ms, input is in s

        self.hw_interface = IPBusDevice
        self.deviceName = name

    def tryStartTransaction(self):

        node = self.hw_interface.getNode(self.deviceName+".Key_Lock")
        lockStatus = node.getClient().rmw_bits(node.getAddress(), 0xffffffff, 0x2); # set lock bit, 0xffffffff is and term, 0x2 is or term
        self.hw_interface.dispatch()

        if (lockStatus & 0x2):
            return False                                                              # device locked

        self.setTimer()

        key = node.read()                                                           # get current key value
        self.hw_interface.dispatch()
        self.accessKey = (key & 0xfffffffc) >> 2

        return True

    def setTimer(self):

        self.hw_interface.getNode(self.deviceName+".Timeout").write(self.fwTimeout)  # set the timeout value
        self.hw_interface.dispatch()

    def isKeyCorrect(self):

        newKey = self.hw_interface.getNode(self.deviceName+".Key_Lock").read()
        self.hw_interface.dispatch()
        newKey = (newKey & 0xfffffffc) >> 2

        if newKey == self.accessKey:
            return True

        return False

    def isErrorBitOn(self):

        regContent = self.hw_interface.getNode(self.deviceName+".Key_Lock").read()
        self.hw_interface.dispatch()

        if (regContent & 0x1):
            return True                                                              # the bit is on

        return False                                                               # all is quiet


    def tryFinishTransaction(self):

        if not self.isKeyCorrect():                       # something went wrong!
            return False

        node = self.hw_interface.getNode(self.deviceName+".Key_Lock")
        node.getClient().rmw_bits(node.getAddress(), 0xfffffffd, 0x0);               # clear the lock bit
        self.hw_interface.dispatch()

        return True


class IpBusInterfaceArbitrationBusy(Exception):
    """
    Exception class to be thrown if IPBus arbitration is busy
    """
    def __init__(self,value):
        self.value = value

    def __str__(self):
        return self.value

class IpBusInterfaceArbitrationError(Exception):
    """
    Exception class to be thrown in there was error in IPBus arbitration
    """
    def __init__(self,value):
        self.value = value

    def __str__(self):
        return self.value


if __name__ == "__main__":


    print "Starting SpiInterface test"


    manager = uhal.ConnectionManager("file:///home/ftmtest/FTMTests/connections_KC705_v6.xml")

    hw = manager.getDevice("Kintex")

    Flash = FlashSpiInterface(hw,"FLASH_SPI_Ram")
    Flash.readID()


    firmware_file = open("files/KC705_v10.bin","rb")
    Flash.validateMemoryWithFile(0x80,firmware_file)

#!/usr/bin/python3
    firmware_file.close


#!/usr/bin/python3

#!/usr/bin/python3

    print "Done!"

