#!/bin/bash
# load ATB fpga from a .bin file
#
# 2 possible installations:
# 1. detector lab installation: ctp3-ipbus installed on flp
# 2. ctp lab installation: ctp3-ipbus installed on a computer different from flp
# See installation variable below and following comments:
#
# Prerequisities:
# - USB on flp connected to ATB's jtag 
# - vivado_lab installed on flp
#   - modify xilinx/vivadorem.sh accordingly
# - trigger@flp:xilinx/load_atb.tcl   -copy of both scripts is in $CTP3_ROOT/scripts/bin/
#               xilinx/vivadorem.sh
#
# 1. detector lab installation: ctp3-ipbus installed on flp
#   Copy scripts to ~/xilinx directory in this computer:
#   mkdir ~/xilinx
#   cd $CTP3_ROOT/scripts/bin
#   cp -a load_atb.tcl vivadorem.sh ~/xilinx/
#   or:

# 2. ctp lab installation: ctp3-ipbus installed on a computer different from flp
#   Arrange passwd-less ssh access to trigger@flp (see flp/flpcon= line further),
#   i.e.: ssh-copy-id trigger@flp
#
#   ssh  trigger@flp 'mkdir ~/xilinx'
#   cd $CTP3_ROOT/scripts/bin
#   scp -p load_atb.tcl vivadorem.sh trigger@flp:xilinx/
#
# Input files:
# - $CTP3_ROOT/boards/atbids.txt
# - $CTP3_ROOT/boards/cfg/xxx.bin     -.bin file to be loaded in atb
#
# Usage:
# loadfpga.sh info
#             -show all possible ATB names with their JTAG ids
# loadfpga.sh atbname ltu_logic_v13_00
# atbname: e.g. atb_15 kcu105 ...
#
# loadfpga.sh jtagid ltu_logic_v13_00
# jtagid: e.g. 210251A535B5 (for atb_9) -can be found at:
#         https://twiki.cern.ch/twiki/bin/view/ALICE/BoardNumbers    
#
function getdigi() {
inpline=`echo $1 | sed 's/ //g'`   # tr -d ' '
IFS='|' read -r a nam c d digi f <<< "$inpline"
n3=${nam:0:3}
if [ -z "$n3" ] ;then
  digi=""
  return
fi
if [ "${digi:0:2}" != "21" ] ;then
  digi=""
  return
fi
#echo $nam $digi =$n3=
#if [ "$n3" = "LM0" ] ;then
if [ "$n3" = "LM0" -o "$n3" = "!LM" -o "$n3" = "ATB" -o "$n3" = "KC7" -o "$n3" = "KCU" ] ;then
  #echo $nam $digi
  return
fi  
}
#
atbids="$CTP3_ROOT/boards/atbids.txt"
if [  -d "/tools/Xilinx/Vivado_Lab"  -a  -e "$HOME/xilinx/vivadorem.sh"  ] ;then
  flp="this computer"
  #flpcon=""
  installation="detlab"
  instext="Vivado lab installed on this computer"
else
  flp='cru-server2'
  flpcon="trigger@$flp"
  installation="ctplab"
  instext="Vivado lab installed on $flp computer"
fi
if [ $# -eq 0 ] ;then
  cat - <<-EOF
Usage ($instext):
loadfpga.sh info         -show all possible ATB names (+ JTAG ids)

Connect USB cable between $flp and ATB board, then
(aan example for LTU 9):

loadfpga.sh atb_9 ltu_logic_v13_00         -load atb_9' fpga, or
loadfpga.sh 210251A535B5 ltu_logic_v13_00  -the same board identified by its JTAGid

Check loadfpga.sh script for more details.

EOF
elif [ "$1" = "info" ] ;then
  # lines=0
  while read line; do
    getdigi "$line"
    if [ -n "$digi" ] ;then
      echo $nam $digi
    fi
    #let lines=lines+1
    #if [ $lines -gt 5 ] ;then
    #  break
    #fi
  done < $atbids
else
  ucname=`echo $1 | tr  [a-z] [A-Z]`
  binname=$2
  if [ ! -e $CTP3_ROOT/boards/cfg/$binname.bin ] ;then
    cat - <<-EOF
Error:
$CTP3_ROOT/boards/cfg/$binname.bin
does not exist.

EOF
  else
    while read line; do
      getdigi "$line"
      if [ "$nam" = "$ucname" -a -n "$digi" ] ;then
        echo $nam $digi
        if [ "$installation" = "ctplab" ] ;then
          scp -p $CTP3_ROOT/boards/cfg/$binname.bin $flpcon:xilinx/
          ssh $flpcon xilinx/vivadorem.sh $digi $binname
        else
          # local copy (if ipbus sw+control on flp):
          ln -sf $CTP3_ROOT/boards/cfg/$binname.bin ~/xilinx/$binname.bin
          vivadorem.sh $digi $binname
        fi
      fi
    done < $atbids
  fi
fi
